/*  byteplex.h - Parallel byte stream module.
    Copyright (c) 2008,  Robert D. Cameron.
    Licensed to the public under the Open Software License 3.0.
    Licensed to International Characters, Inc., under the Academic
    Free License 3.0.

    This module has as its goal the buffering of XML byte data and
    transformation of 16-bit and 32-bit code unit data so that the
    parsing engine is provided a uniform representation based on
    the concept of an 8-bit pseudo-ASCII representation (x8data).

    A Byteplex object provides buffers for one to six parallel data
    streams based for an XML input entity, depending on the size of
    character code units.  
       1.  In the case of 8-bit code units, a single byte stream
           consisting of unmodified input data is maintained.
           x8data = src_buffer
       2.  In the case of 16-bit code units (UTF-16 and UCS-2 families),
             (a) the original code unit stream is maintained unmodified,
             (b) the x16hi byte stream is established for the high byte
                 of each code unit,
             (c) the x16lo byte stream is established for the low byte
                 of each code unit, and
             (d) x8data is established as the pseudo-ASCII byte stream,
                 with ASCII code units having their proper 8-bit values,
                 and all others having bit 0 set to 1.
       3.  In the case of 32-bit code units (UTF-32 family),
             (a) the original code unit stream is maintained unmodified,
             (b) the x32hh byte stream has high bytes of each code unit
             (c) the x32hl byte stream has second bytes of each code unit
             (d) the x32lh byte stream has third bytes of each code unit
             (e) the x32hh byte stream has low bytes of each code unit, and
             (f) x8data is established as the pseudo-ASCII byte stream,
                 with ASCII code units having their proper 8-bit values,
                 and all others having bit 0 set to 1.

    The pseudo-ASCII representation is defined for both ASCII-based
    and EBCDIC-based character sets such that all characters in
    the ASCII repertoire (i.e., having Unicode code points from 0x00
    to 0x7F), are represented as themselves and no non-ASCII character
    is represented as a character in the ASCII repertoire.

*/

#ifndef BYTEPLEX_H
#define BYTEPLEX_H

#include "xmldecl.h"
#include "../lib/lib_simd.h"

/* The BytePack and the BitBlock are the two fundamental
   types used by the parabix program for data held in 
   SIMD registers, representing, respectively, the byte-oriented
   and bit-oriented views of character data.*/

typedef SIMD_type BytePack;
typedef SIMD_type BitBlock;
const int PACKSIZE = sizeof(SIMD_type);
const int BLOCKSIZE = sizeof(SIMD_type) * 8;

/* Define the size of buffer used for lexical analysis/parsing. */
const int BUFFER_BLOCKS = 781;
const int BUFFER_SIZE = BUFFER_BLOCKS * BLOCKSIZE;

/* When working near the end of a buffer, a bytespace test may involve
   a multibyte literal.  The bytespace buffer must always make available
   a number of lookahead bytes at least equal to the maximum length of any
   such literal. */

const int LOOKAHEAD_POSITIONS = 16;
const int BYTEPLEX_SIZE = BUFFER_SIZE + LOOKAHEAD_POSITIONS;

class Byteplex {
public:
	virtual ~Byteplex();
	static Byteplex * ByteplexFactory(Entity_Info * e);
	static Byteplex * ByteplexFactory(Entity_Info * e, FILE * inputfile);
	static Byteplex * ByteplexFactory(Entity_Info * e, unsigned char * buffer_bytes, int buffer_size);
	virtual void DoByteplex() = 0;
	virtual void PreparePseudoASCII_Stream() = 0;
	virtual	void InitializeBuffer(unsigned char * src, int lgth) = 0;
	virtual void AdvanceInputBuffer(int advance_amt) = 0;
	virtual int UTF8_Length(int name_pos, int lgth)=0;
	virtual void to_UTF8(int name_pos, int lgth, char * u8_ptr)=0;
	/* Source code unit buffer. */
	BytePack * src_buffer;
	int units_in_buffer;

	/* Pseudo-ASCII stream. */
	BytePack * x8data;

protected:
	FILE * infile;
	int packs_in_buffer;
	int CopyAndFill(unsigned char * bytes_to_copy, int lgth, int bytes_to_read);
	void Set_limits(int units_in_buffer);

};


/*  The X8_Buffer template class is used for either ASCII- or EBCDIC-
    based 8-bit code units.
    The X8_Buffer<ASCII> class includes 7-bit ASCII 
    (with high-order bit 0), the ISO-8859 character sets and UTF-8.

    The family of 8-bit EBCDIC based character sets are processed using
    the X8_Buffer<EBCDIC> class.
*/

template <CodeUnit_Base C>
class X8_Buffer : public Byteplex {
public:
	static const CodeUnit_Base Base = C;
	static const CodeUnit_Size Size = SingleByte;
	X8_Buffer();
	virtual ~X8_Buffer();

	void DoByteplex();
	void PreparePseudoASCII_Stream();
	void AdvanceInputBuffer(int advance_amt);
	void InitializeBuffer(unsigned char * src, int lgth);
	int UTF8_Length(int name_pos, int lgth);
	void to_UTF8(int name_pos, int lgth, char * u8_ptr);
};

class UTF8_Buffer : public Byteplex {
public:
	static const CodeUnit_Base Base = ASCII;
	static const CodeUnit_Size Size = SingleByte;
	UTF8_Buffer();
	virtual ~UTF8_Buffer();

	void DoByteplex();
	void PreparePseudoASCII_Stream();
	void AdvanceInputBuffer(int advance_amt);
	void InitializeBuffer(unsigned char * src, int lgth);
	int UTF8_Length(int name_pos, int lgth);
	void to_UTF8(int name_pos, int lgth, char * u8_ptr);
};


/*  UTF-16 and UCS-2 character set families in BE and LE byte orders. 
    The U16LE and U16BE subclasses each provide a distinct byteplexer to 
    produce 2 parallel byte streams for the high and low bytes of each
    16-bit code unit.  Once byteplexing is complete, a generic pseudoASCII 
    conversion routine can be applied at the U16_Buffer level. */

class U16_Buffer : public Byteplex {
public:
	static const CodeUnit_Base Base = ASCII;
	static const CodeUnit_Size Size = DoubleByte;
	U16_Buffer();
	virtual ~U16_Buffer();
	virtual void DoByteplex() = 0;
	void PreparePseudoASCII_Stream();
	void AdvanceInputBuffer(int advance_amt);
	void Validate_UTF16();
	void Validate_UCS2();
	void InitializeBuffer(unsigned char * src, int lgth);
	int UTF8_Length(int name_pos, int lgth);
	void to_UTF8(int name_pos, int lgth, char * u8_ptr);
protected:
	BytePack * x16hi;
	BytePack * x16lo;
};

class U16LE_Buffer : public U16_Buffer {
public:
	U16LE_Buffer();
	void DoByteplex();
};

class U16BE_Buffer : public U16_Buffer {
public:
	U16BE_Buffer();
	void DoByteplex();
};


/*  UTF-32/UCS-4 character sets in BE, LE, 2143 and 3412 byte orders. 
    Each subclass of U32_Buffer provide a distinct byteplexer to 
    produce the 4 parallel byte streams of Unicode data.  Once
    byteplexing is complete, a generic pseudoASCII routine can
    be applied. */
class U32_Buffer : public Byteplex {
public:
	static const CodeUnit_Base Base = ASCII;
	static const CodeUnit_Size Size = QuadByte;
	U32_Buffer();
	virtual ~U32_Buffer();
	virtual void DoByteplex() = 0;
	void PreparePseudoASCII_Stream();
	void AdvanceInputBuffer(int advance_amt);
	void Validate_UTF32();
	void InitializeBuffer(unsigned char * src, int lgth);
	int UTF8_Length(int name_pos, int lgth);
	void to_UTF8(int name_pos, int lgth, char * u8_ptr);
protected:
	BytePack * x32hh;
	BytePack * x32hl;
	BytePack * x32lh;
	BytePack * x32ll;
};

class U32LE_Buffer : public U32_Buffer {
public:
	U32LE_Buffer();
	void DoByteplex();
};

class U32BE_Buffer : public U32_Buffer {
public:
	U32BE_Buffer();
	void DoByteplex();
};

class U32_2143_Buffer : public U32_Buffer {
public:
	U32_2143_Buffer();
	void DoByteplex();
};

class U32_3412_Buffer : public U32_Buffer {
public:
	U32_3412_Buffer();
	void DoByteplex();
};


inline char * copy_name (char * s, int lgth){		
	char * d = new char[lgth+1];
	memcpy(d, s,lgth); 
	d[lgth] = '\0';	
	return d;
}

#endif


